(** S2DLDraws
@Name: Simple 2D Library - Main Surface Drawing Object
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.11
@Note: Do not initialize yourself !!! - Automatically done !!!
@Change_History:
@   ver 1.11 - Working on Linux (Daniel Mantione)
@   ver 1.10 - Added Fading effects
@            - Added Scene Management
@   ver 1.04 - Added Cursor Support
@   ver 1.03 - Added Flip to S2DLDraws to allow calculation of FPS
@           - Auto check a config file to see if must run full screen.
@Copyright: (c) 2004-2007 by William Cairns (CairnsGames)

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGDev web site.

@Usage: Just include this unit as the first S2DL Unit in the Main Program
   S2DLDraws is responsible for setting up the required SDL surface from the Screen
   S2DLDraws will also end the SDL session and clear its own surface

@Author: W.Cairns **)
unit S2DLDraws;

{$ifdef FPC}
{$mode objfpc}
{$endif}
{$H+}

interface


Uses
  Classes,
  SysUtils,
  S2DLLogger,
  gl,
  glu,
  SDL,
  S2DLImage;

Const
  fNoFade = 0; fIn = 1; fOut = 2;

Type
  (**
    TS2DLDraw encapsulated the screen area for the SDL application. A single
    instance of the class is created and used to control the display area.
    Methods and properties are also included to calculate the Frame rate
    currently being seen in the game.
    @
    This class should never be instantiated as it is done automatically in
    the initialization section of the unit.
  **)
  TS2DLDraw = Class(TObject)
  Private
    FSurface : PSDL_Surface;
    LastTick : UInt32;
    LastFPS : Integer;
    CurFPS : Integer;
    FAppPath: String;
    FR,FG,FB,FA : Single;
    FadeTick : Cardinal;
    FadeSecs : Single;
    HasCursor : Boolean;
    CursorImage : TS2DLImage;
    procedure SetAppPath(const Value: String);
  Public
    FadeMode : Integer;
    Cursor : Boolean;
    CursorX, CursorY, CursorIndex : Integer;
    Property Surface : PSDL_Surface read FSurface;
    Property AppPath : String read FAppPath write SetAppPath;
    Constructor Create;
    Destructor Destroy; Override;
    Procedure Clear;
    Procedure Flip;
    Procedure ScreenColor(R : Single = -1; G : Single = -1; B : Single = -1; A : Single = -1);
    Procedure SetColor(R,G,B : Single);
    Procedure SetColor(R,G,B,A : Single);
    Procedure TempColor(R,G,B : Single);
    Procedure ClearColor(R,G,B : Single);
    Procedure FadeIn(Secs : Single = 1);
    Procedure FadeOut(Secs : Single = 1);
    Procedure SceneStart;
    Procedure SceneEnd;
    Procedure LoadCursor(CursorName : String);
    Function FPS : Integer;
    Procedure WindowCaption(NewCaption : String);
    Procedure WindowResize(W,H:integer);
  End;

Var
  (**S2DLDraw is instantiated by the S2DL code and can be used in the
  program.**)
  S2DLDraw : TS2DLDraw;
  Error : Boolean;
  Keys : Array[0..SDLK_EURO] of Boolean;

Function MakeSDLColor(R,G,B : UInt8) : TSDL_Color;

implementation

uses glext;

(**
  Create an TSDL_Color from the basic colors

  @param   R as an UInt8
  @param   G as an UInt8
  @param   B as an UInt8
  @return  a TSDL_Color
**)
Function MakeSDLColor(R,G,B : UInt8) : TSDL_Color;
Begin
  Result.R := R;
  Result.G := G;
  Result.B := B;
End;

{ TS2DLDraw }


(**
  This is the constructor method for the TS2DLDraw class.
**)
Constructor TS2DLDraw.Create;
Var
  videoInfo : PSDL_VideoInfo;
    videoFlags: cardinal;
//    fSurface:PSDL_Surface; // no need to make it public
Begin
  LastFPS := 0;
  CurFPS := 0;
  FR := 1; FG := 1; FB := 1; FA:=1;
  Cursor := False; HasCursor := False;
  Error := false;
  if SDL_Init(SDL_INIT_VIDEO) = -1 then
  begin
    Log.LogError(Format('Could not initialize SDL : %s',
      [SDL_GetError]), 'S2DLDraw.Create');
    SDL_Quit;
    halt(1);
  end;

  // To set up full screen mode add
  //     or SDL_FULLSCREEN
{  If FullScreen then
    Flags := SDL_SWSURFACE or SDL_HWPALETTE or SDL_FULLSCREEN
  Else
    Flags := SDL_SWSURFACE or SDL_HWPALETTE;

  FSurface := SDL_SetVideoMode(800, 600, 24, Flags);
  if FSurface = nil then
  begin
    Log.LogError(Format('Couldn''t set 800x600x24 video mode : %s',[SDL_GetError]), 'S2DLDraw.Create');
    Error := true;
  end;

  Rect.X := 0;
  Rect.Y := 0;
  Rect.W := FSurface.W;
  Rect.H := FSurface.H;
  LastTick := SDL_GetTicks;
  FAppPath := ExtractFilePath( ParamStr(0) );}

{$ifdef win32}
  if wglSwapIntervalEXT<>nil then
    wglSwapIntervalEXT(1);
{$endif}

  // Fetch the video info
  videoInfo := SDL_GetVideoInfo;

  if ( videoInfo = nil ) then
  begin
//    result:=false;
    exit;
  end;

  // the flags to pass to SDL_SetVideoMode
  videoFlags := $00000002;                  // Enable OpenGL in SDL
  videoFlags := videoFlags or SDL_DOUBLEBUF; // Enable double buffering
  videoFlags := videoFlags or SDL_HWPALETTE; // Store the palette in hardware
  videoFlags := videoFlags or SDL_RESIZABLE;
  if fullscreen then
    videoFlags := videoFlags or SDL_FULLSCREEN;

  // This checks and see if surfaces can be stored in memory
  if ( videoInfo^.hw_available <> 0 ) then
    videoFlags := videoFlags or SDL_HWSURFACE
  else
    videoFlags := videoFlags or SDL_SWSURFACE;

  // This checks if hardware blits can be done * /
  if ( videoInfo^.blit_hw <> 0 ) then
    videoFlags := videoFlags or SDL_HWACCEL;

  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
  fsurface := SDL_SetVideoMode(800,600, 32, videoflags );

  if ( surface = nil ) then
  begin
//    result:=false;
    exit;
  end;


  SDL_ShowCursor(0);

  // opengl initializations
  // ----------------------

  // Enable smooth shading
  glShadeModel( GL_SMOOTH );

  // Set the background black
  glClearColor( 0.0, 0.0, 0.0, 0.0 );

  // Disables Depth Testing (no z buffer)
  glDisable( GL_DEPTH_TEST );
  glDisable(GL_CULL_FACE);

  // Really Nice Perspective Calculations
  glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

  glViewport(0, 0, 800 , 600);

  // orthographic setup
  
   // go to matrix proj
   glMatrixMode(GL_PROJECTION);
   // reset and...
   glLoadIdentity();
   // ...set our ortho
   glOrtho(0, 800, 600, 0, 0, 100);
   // return to modelview, and reset it
   glMatrixMode(GL_MODELVIEW);

   glLoadIdentity();

   // activate textures!
   glEnable(GL_TEXTURE_2D);
  
   // this is the blend function for alpha blending.
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   AppPath := ExtractFilePath(ParamStr(0));
End;


(**
  This is the destructor method for the TS2DLDraw class.
**)
Destructor TS2DLDraw.Destroy;
Begin
  Log.LogStatus('S2DLDraw surface being cleared','TS2DLDraw.Destroy');
  SDL_FreeSurface(FSurface);
End;

(**
  Clear the SDL Surface
**)
procedure TS2DLDraw.Clear;
begin
//  SDL_FillRect(FSurface, @Rect, 0);
  glClear( GL_COLOR_BUFFER_BIT);

  // not really useful since we never change it
  glLoadIdentity;
end;

(**
  Flips the SDL surface and does the necessary calculations to calculate the
  current FPS.
**)
procedure TS2DLDraw.Flip;
begin
  CurFPS := CurFPS + 1;
  If SDL_GetTicks > LastTick + 1000 then
    Begin
      LastFPS := CurFPS;
      CurFPS := 0;
      LastTick := SDL_GetTicks;
      Try
        S2DLDraw.WindowCaption('FPS: '+IntToStr(S2DLDraw.FPS));
      Except
        On E:Exception do
          Begin
            Log.LogError(E.Message,'Render');
          End;
      End;
    End;
  SDL_GL_SwapBuffers;
end;

(**Returns the current Frame Per Second calculation. 

  @return  an Integer
**)
function TS2DLDraw.FPS: Integer;
begin
  If LastFPS = 0 then
    Begin
      If CurFPS = 0 then
        Result := 1
      Else
        Result := CurFPS
    End
  Else
    Result := LastFPS;
end;

(**
  Set the screen caption for the windowed window.

  @param   NewCaption as a String
**)
procedure TS2DLDraw.WindowCaption(NewCaption: String);
begin
  SDL_WM_SetCaption(PChar(NewCaption), nil);
end;

(**
  This is a setter method for the AppPath property. This is used to define some
  settings such as the location of Logs etc.

  @param   Value as a String constant
**)
procedure TS2DLDraw.SetAppPath(const Value: String);
begin
  FAppPath := Value;
end;

(**
  Set the Global screen color. This change is temporary.

  @param   R as a Single , use -1 for current Global color setting.
  @param   G as a Single, use -1 for current Global color setting.
  @param   B as a Single, use -1 for current Global color setting.
**)
procedure TS2DLDraw.ScreenColor(R, G, B, A: Single);
begin
  If R = -1 then R := FR;
  If G = -1 then G := FG;
  If B = -1 then B := FB;
  If A = -1 then A := FA;
  glColor4f(R,G,B,A);
end;

(**
  End the scene and flip. If sursor is on and loaded it will be drawn before flipping.
**)
procedure TS2DLDraw.SceneEnd;
begin
  If Cursor then
    CursorImage.Draw(CursorX, CursorY, CursorIndex);
  Flip;
end;

(**
  Start a new scene. Will manage the relevant screen fading in/out and sets
  the global color scheme.
  Note the Temporary color changes can overwrite the Global sceme unless ended
  after being used to draw.
**)
procedure TS2DLDraw.SceneStart;
Var
  TC : Integer;
  MS : Single;
begin
  If FadeMode > fNoFade then
    Begin
      TC := SDL_GetTicks - FadeTick;
      MS := TC / (1000 * FadeSecs);
      If FadeMode = fOut then
        MS := -MS;
      FR := FR + MS;
      FG := FG + MS;
      FB := FB + MS;
      FA := 1;
      FadeTick := SDL_GetTicks;
      If (FR < 0) or (FR > 1) then
        FadeMode := fNoFade;
    End;
  ScreenColor(FR,FG,FB);
end;

(**
  This is a setter method for the Color property. Sets the Global values.

  @param   R as a Single
  @param   G as a Single
  @param   B as a Single
**)
procedure TS2DLDraw.SetColor(R, G, B: Single);
begin
  FR := R;
  FG := G;
  FB := B;
  FA := 1;
  ScreenColor;
end;

procedure TS2DLDraw.SetColor(R, G, B, A: Single);
begin
  FR := R;
  FG := G;
  FB := B;
  FA := A;
  ScreenColor;
end;

(**
  Start scene fadein taking Secs Seconds.

  @param   Secs as a Single
**)
procedure TS2DLDraw.FadeIn(Secs: Single);
begin
  FadeMode := fIn;
  FR := 0; FG := 0; FB := 0; FA := 1;
  FadeSecs := Secs;
  FadeTick := SDL_GetTicks;
end;

(**
  Start Scene FadeOut - taking Secs seconds.

  @param   Secs as a Single
**)
procedure TS2DLDraw.FadeOut(Secs: Single);
begin
  FadeMode := fOut;
  FR := 1; FG := 1; FB := 1; FA := 1;
  FadeSecs := Secs;
  FadeTick := SDL_GetTicks;
end;

(**
  Load a cursor. The Cursor is then marked as visible.

  @param   CursorName as a String
**)
procedure TS2DLDraw.LoadCursor(CursorName: String);
begin
  HasCursor := True;
  Cursor := True;
  CursorImage := TS2DLImage.Create;
  CursorImage.LoadFromFile(CursorName);
  CursorImage.Transparent := True;
  CursorImage.PatternWidth := 64;
  CursorImage.PatternHeight := 64;
end;

(**
  Temporarily set the colors. These do not affect the current global settings.

  @param   R as a Single
  @param   G as a Single
  @param   B as a Single
**)
procedure TS2DLDraw.TempColor(R, G, B: Single);
begin
  glColor3f(R,G,B);
end;

(**
  Set the clear colour.

  @param   R as a Single
  @param   G as a Single
  @param   B as a Single
**)
procedure TS2DLDraw.ClearColor(R, G, B: Single);
begin
  glClearColor(R,G,B,0.0);
end;

procedure TS2DLDraw.WindowResize(W,H:integer);

var videoflags:cardinal;
    videoInfo : PSDL_VideoInfo;

begin
  videoInfo := SDL_GetVideoInfo;

  if ( videoInfo = nil ) then
    raise exception.create('Cannot set video mode');
  // the flags to pass to SDL_SetVideoMode
  videoFlags := $00000002;                  // Enable OpenGL in SDL
  videoFlags := videoFlags or SDL_DOUBLEBUF; // Enable double buffering
  videoFlags := videoFlags or SDL_HWPALETTE; // Store the palette in hardware
  videoFlags := videoFlags or SDL_RESIZABLE;
  if fullscreen then
    videoFlags := videoFlags or SDL_FULLSCREEN;

  // This checks and see if surfaces can be stored in memory
  if ( videoInfo^.hw_available <> 0 ) then
    videoFlags := videoFlags or SDL_HWSURFACE
  else
    videoFlags := videoFlags or SDL_SWSURFACE;

  // This checks if hardware blits can be done * /
  if ( videoInfo^.blit_hw <> 0 ) then
    videoFlags := videoFlags or SDL_HWACCEL;
  fsurface := SDL_SetVideoMode(W,H, 32, videoflags );
  if fsurface=nil then
    raise exception.create('Cannot set video mode');
  glViewport(0, 0, W , H);
end;

Initialization
  // This initalizes the SDL session and creates the SDL surface for Drawing
  Log.LogStatus('SDL being started up.','S2DLDraws.Initialization');
  S2DLDraw := TS2DLDraw.Create;

Finalization
  // This shut downs the SDL session and will fre the SDL surface
  Log.LogStatus('SDL being closed down.','S2DLDraws.Finalization');
  S2DLDraw.Free;
  SDL_Quit;

end.
