(** S2DLImage
@Name: Simple 2D Library - Images Unit
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.11
@Change_History:
@  ver 1.11 - Working on Linux (Daniel Mantione)
@  ver 1.10 - Removed integration with Windows GDI functions through Bitmap
@           - Removed Legacy Surface functions
@           - Added DrawStretch  funtions
@           - Added DrawRotate funtions
@           - Added DrawAlpha funtions
@  ver 1.05 - Support for OpenGL through SDL
@           - Hidden GDI support due to file size increase.
@  ver 1.04 - Added integration with Windows GDI functions through Bitmap
@           - Modified to Load BMP, JPG, GIF etc

@Copyright: (c) 2004-2007 by William Cairns (CairnsGames)

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGD web site.

@Usage: A single S2DLImageList is created by default at the start of the
  program. This imagelist can be accessed by Images.<method>  @
  Additional S2DLImageLists can be created if required.
@Note: Do not initialize Images:TS2DLImageList yourself !!! - Automatically done !!!

@Author: W.Cairns **)
Unit S2DLImage;

{$ifdef FPC}
{$mode delphi}
{$endif}

Interface
{ TODO : Move Copy Image into seperate unit (A utils unit or something) }               

Uses
  Classes,
  SysUtils,
  S2DLLogger,
  gl,
  glu,
  textureunit,
  sdl,
  Variants;

Type
  (** Forward **)
  TS2DLImage = Class;
  (**
    TS2DLImageList is an image repository where images can be stored and accessed
    as needed in the program. Each image stored in the list is stored as a
    TS2DLImage. Includes functionality to @ - Search @ - Delete @ - Load Images 
  **)
  TS2DLImageList = Class
    Private
      FItems : TList;
      Function GetImage(Index : Integer) : TS2DLImage;
      Function GetImageByName(Name : String) : TS2DLImage;
      function DefaultGetAnImage(Index: Variant): TS2DLImage;
    Public
      Property Items[Index : Integer] : TS2DLImage read GetImage;
      Property ItemByName[Name : String] : TS2DLImage read GetImageByName;
      Property GetAnImage[Index : Variant] : TS2DLImage read DefaultGetAnImage; default;
      Constructor Create;
      Destructor Destroy; Override;
      Function Find(Name : String) : TS2DLImage;
      Function FindID(Name : String) : Integer;
      Procedure Add(Item : TS2DLImage);
      Function Count : Integer;
      Function Delete(Item : Integer) : Boolean;
      Procedure LoadImage(FileName : String); Overload;
      Procedure LoadImage(FileName : String; Name : String); Overload;
      Procedure LoadImage(FileName : String; Name : String; IsTransparent : Boolean); Overload;
      Function LoadFromFile(FileName : String; Name : String) : TS2DLImage;
  End;
  (**
    TS2DLImage encapsulates the SDL surface functionality into a class that
    can more easily be used in a program. TS2DLImage is used as an image
    respository and called to draw the image onto the screen.
  **)
  TS2DLImage = Class
  private
    FPatternWidth: Integer;
    FPatternHeight: Integer;
    function GetHeight: Integer;
    function GetWidth: Integer;
    procedure SetPatternHeight(const Value: Integer);
    procedure SetPatternWidth(const Value: Integer);
  public
    Name : String;
    ImageSurface: PSDL_Surface;
    Texture : TTexture;
    SrcRect, DestRect: TSDL_Rect;
    PatternWidthFrac, PatternHeightFrac : Single;
    Transparent : Boolean;
    Property Width : Integer read GetWidth;
    Property Height : Integer read GetHeight;
    Property PatternWidth : Integer read FPatternWidth write SetPatternWidth;
    Property PatternHeight : Integer read FPatternHeight write SetPatternHeight;
    Constructor Create;
    Constructor CreateSizedSurface(Width,Height,ColorDepth : Integer);
    Destructor Destroy; Override;
    Procedure LoadFromFile(FileName : PChar); Overload;
    Procedure LoadFromFile(FileName : String); Overload;
    Procedure LoadFromFile(FileName : String; IsTransparent : Boolean); Overload;
    Procedure DrawRect(X, Y, W, H: Integer; X1, Y1, X2, Y2: Single);
    Procedure DrawStretch(X,Y,W,H : Integer); Overload;
    Procedure DrawStretch(X,Y,W,H, Pattern : Integer); Overload;
    Procedure DrawStretchRotate(X,Y,W,H, Pattern,Degrees : Integer); Overload;
    Procedure DrawStretchAlpha(X,Y,W,H, Pattern : Integer; Alpha : Single); Overload;
    Procedure DrawStretchRGBA(X,Y,W,H:integer;R,G,B,A:single); Overload;
    Procedure Draw(X,Y : Integer); Overload;
    Procedure Draw(X,Y : Integer; Pattern : Integer); Overload;
    Procedure DrawRotate(X,Y : Integer; Angle : Single); Overload;
    Procedure DrawRotate(X,Y : Integer; Pattern : Integer; Angle : Single); Overload;
    Procedure DrawAlpha(X,Y : Integer; Alpha : Single); Overload;
    Procedure DrawAlpha(X,Y : Integer; Pattern : Integer; Alpha : Single); Overload;
    Procedure DrawTiled(X,Y,W,H : Integer); Overload;
    Procedure DrawTiled(X,Y,W,H,Pattern : Integer); Overload;
    Procedure DrawArea(X,Y : Integer; IX, IY, IW, IH : Integer);
    Procedure SetPixel(X,Y : Integer; Color : UInt32);
//    Procedure CopyBitmap(X,Y : Integer; BMP : TBitmap; TransparentColor : UInt32);
    Procedure BindImage;
  End;

Var
  Images : TS2DLImageList;

Implementation


Uses
  S2DLDraws;

{ TS2DLImageList }

(**
  This is the constructor method for the TS2DLImageList class. Creates the list to
  store the images in.
**)
Constructor TS2DLImageList.Create;
Begin
  FItems := TList.Create;
End;

(**
  This is the destructor method for the TS2DLImageList class. Clear memory
  of all loaded images.
**)
Destructor TS2DLImageList.Destroy;
Var
  I : Integer;
Begin
  // Free memory used for each image
  For I := FItems.Count - 1 downto 0 do
    Begin
      TS2DLImage(FItems[I]).Free;
      FItems.Delete(I);
    End;
  FItems.Free;
End;

(**
  This is a getter method for the Image property.

  @param   Index as an Integer
  @return  a TS2DLImage
**)
Function TS2DLImageList.GetImage(Index : Integer) : TS2DLImage;
Begin
  Result := TS2DLImage(FItems[Index]);
  If Result = Nil then
    Log.LogError('Image with index:'+IntToStr(Index)+' not found','TS2DLImageList.GetImage');
End;

(**
  This is a getter method for the ImageByName property.

  @param   Name as a String
  @return  a TS2DLImage
**)
Function TS2DLImageList.GetImageByName(Name : String) : TS2DLImage;
Var
  I : Integer;
Begin
  Result := Nil;
  Try
    For I := FItems.Count-1 downto 0 do
      Begin
        If TS2DLImage(FItems[I]).Name = Name then
          Begin
            Result := TS2DLImage(FItems[I]);
            Exit;
          End;
      End;
  except
    On E:Exception do
      Begin
        Log.LogError('Cannot find Image:'+Name,'TS2DLImageList.GetImageByName');
      End;
  end;
End;

(**
  Find an image based on its Name.

  @param   Name as a String
  @return  a TS2DLImage
**)
Function TS2DLImageList.Find(Name : String) : TS2DLImage;
Begin
  Result := GetImageByName(Name);
End;

(**
  Adds a new image into the Images List.

  @param   Item as a TS2DLImage
**)
Procedure TS2DLImageList.Add(Item : TS2DLImage);
Begin
  FItems.Add(Item);
End;

(**
  Returns number of images stored in the image list.

  @return  an Integer
**)
Function TS2DLImageList.Count : Integer;
Begin
  Result := FItems.Count;
End;

(**
  Removes and frees an image from the list.

  @param   Item as an Integer
  @return  a Boolean
**)
Function TS2DLImageList.Delete(Item : Integer) : Boolean;
Var
  GameImage : TS2DLImage;
Begin
  GameImage := TS2DLImage(FItems[Item]);
  FItems.Delete(Item);
  GameImage.Free;
  Result := True;
End;

(**
  Loads an image. The filename must be either absolute of based on current
  EXE location.

  @param   Name     as a String
  @param   FileName as a String
**)
procedure TS2DLImageList.LoadImage(FileName: String; Name: String);
Begin
  LoadFromFile(FileName,Name);
end;


(**
  Loads an image. The filename must be either absolute of based on current
  EXE location.
  The Image is automatically added to the ImageList

  @param   Name     as a String
  @param   FileName as a String
  @param   R        as an Integer
  @param   G        as an Integer
  @param   B        as an Integer
**)
procedure TS2DLImageList.LoadImage(FileName, Name: String; IsTransparent : Boolean);
Var
  Image : TS2DLImage;
begin
  Image := LoadFromFile(FileName,Name);
  Image.Transparent := IsTransparent;
end;

(**
  Find an Image based on its name. But returns the index number of the image in the list.

  @param   Name as a String
  @return  an Integer
**)
function TS2DLImageList.FindID(Name: String): Integer;
Var
  I : Integer;
Begin
  For I := FItems.Count-1 downto 0 do
    Begin
      If TS2DLImage(FItems[I]).Name = Name then
        Begin
          Result := I;
          Exit;
        End;
    End;
  Result := -1;
end;

{ TS2DLImage }

(**
  This is a getter method for the Height property. Returns effective height
  used by the image when drawing - if Patetrns are set returns PatternHeight,
  else returns Surface Height.

  @return  an Integer
**)
function TS2DLImage.GetHeight: Integer;
begin
  If PatternHeight > 0 then
    Result := PatternHeight
  Else
    Result := Texture.Height;
end;

(**

  This is a getter method for the Width property. Returns effective Width
  used by the image when drawing - if Patetrns are set returns PatternWidth,
  else returns Surface Width.

  @return  an Integer

**)
function TS2DLImage.GetWidth: Integer;
begin
  If PatternWidth > 0 then
    Result := PatternWidth
  Else
    Result := Texture.Width;
end;

(**

  This is the constructor method for the TS2DLImage class.

**)
Constructor TS2DLImage.Create;
Begin
End;

(**
  This is the constructor method for the TS2DLImage class. It creates a surface with
  size and color depth defined. This is useful for creating a temporary image
  that a background or other composite image can be created on.

  @param   Width      as an Integer
  @param   Height     as an Integer
  @param   ColorDepth as an Integer
**)
Constructor TS2DLImage.CreateSizedSurface(Width,Height,ColorDepth : Integer);
Begin
  ImageSurface := SDL_CreateRGBSurface(SDL_SWSURFACE or SDL_HWPALETTE, Width,Height, ColorDepth, 0, 0, 0, 0);
End;

(**
  This is the destructor method for the TS2DLImage class. Frees the memory
  allocated to the image surface.
**)
Destructor TS2DLImage.Destroy;
Begin
//  If Assigned(ImageSurface) then
//    SDL_FreeSurface(ImageSurface);
End;

(**
  Loads an image. The filename must be either absolute of based on current
  EXE location.

  @param   FileName as a PChar
**)
Procedure TS2DLImage.LoadFromFile(FileName : PChar);
Begin
  If FileExists(FileName) then
    Begin
      Texture := TTexture.Create;
      Texture.loadFrom(FileName);
    End
  Else
    Begin
      Log.LogError('File:'+FileName+' not found','TS2DLImage.LoadFromFile');
    End;
End;

(**
   Draw the current texture to the screen.

  @param   ox as a single
  @param   oy as a single
  @param   ex as a single
  @param   ey as a single
  @param   x  as a single
  @param   y  as a single
  @param   w  as a single
  @param   h  as a single
**)
procedure DrawRectangle(ox,oy,ex,ey,x,y,w,h:single); Overload;
begin

  glBegin( GL_QUADS  );         // Drawing Using Triangles
    glTexCoord2f(ox,oy);
    glVertex2f(  x,  y );

    glTexCoord2f(ex,oy);
    glVertex2f(  x+W,  y );

    glTexCoord2f(ex,ey);
    glVertex2f(  x+W,  y+H );

    glTexCoord2f(ox,ey);
    glVertex2f(  x,  y+H );

  glEnd;                            // Finished Drawing The Triangle
end;
(**
   Draw the current texture to the screen.

  @param   ox as a single - Portion of Image to Draw
  @param   oy as a single
  @param   ex as a single
  @param   ey as a single
  @param   x  as a single - Location of Image to Draw
  @param   y  as a single
  @param   w  as a single
  @param   h  as a single
  @param   a  as a single - Degrees

**)
procedure DrawRectangle(ox,oy,ex,ey,x,y,w,h,a:single); Overload;
begin
  glPushMatrix;
  glTranslatef(x+W/2,y+H/2,0);  //transate to the center of the image
  glRotatef(A,0,0,1); //rotate
  glBegin( GL_TRIANGLE_STRIP  );         // Drawing Using Triangles
    glTexCoord2f(ox,oy);
    glVertex2f( -w / 2, -h / 2);
    glTexCoord2f(ox,ey);
    glVertex2f( -w / 2, +H / 2);
    glTexCoord2f(ex,oy);
    glVertex2f( +W / 2, -h / 2);
    glTexCoord2f(ex,ey);
    glVertex2f( +W / 2, +H / 2);
  glEnd;                            // Finished Drawing The Triangle
  glPopMatrix;
end;
(**
   Draw the current texture to the screen.

  @param   x as a single
  @param   y as a single
  @param   w as a single
  @param   h as a single
**)
procedure DrawRectangle(x,y,w,h:single); Overload;
begin
  glBegin( GL_TRIANGLE_STRIP  );         // Drawing Using Triangles
    glTexCoord2f(0,0);
    glVertex2f(  x,  y );

    glTexCoord2f(0,1);
    glVertex2f(  x,  y+H );

    glTexCoord2f(1,0);
    glVertex2f(  x+W,  y );

    glTexCoord2f(1,1);
    glVertex2f(  x+W,  y+H );
  glEnd;                            // Finished Drawing The Triangle
end;

(**
   Draw the current texture to the screen.

  @param   x as a single
  @param   y as a single
  @param   w as a single
  @param   h as a single
  @param   a as a single - Angle to Rotate Object
**)
procedure DrawRectangle(x,y,w,h,a:single); Overload;
begin
  glPushMatrix;
  glTranslatef(x+W/2,y+H/2,0);  //transate to the center of the image
  glRotatef(A,0,0,1); //rotate
  glBegin( GL_TRIANGLE_STRIP  );         // Drawing Using Triangles
    glTexCoord2f(0,0);
    glVertex2f( -w / 2, -h / 2);
    glTexCoord2f(0,1);
    glVertex2f( -w / 2, +H / 2);
    glTexCoord2f(1,0);
    glVertex2f( +W / 2, -h / 2);
    glTexCoord2f(1,1);
    glVertex2f( +W / 2, +H / 2);
  glEnd;                            // Finished Drawing The Triangle
  glPopMatrix;
end;

(**
  Draws the Surface onto the Screen. Has been overridden to support multiple
  Different parameter collections. @ This implementation of draw enables the
  display of a section of the image at position X,Y of the S2DLDraw surface. The
  section of the image to display is defined as the pattern number. The number
  of patterns in an image is defined by the PatternHeight and PatternWidth
  properties.

  @param   X        as an Integer
  @param   Y        as an Integer
**)
Procedure TS2DLImage.Draw(X,Y : Integer);
Begin
  If Transparent then
    Begin
      glEnable(GL_BLEND);
//      glColor4f( 1.0, 1.0, 1.0, 1.0);
//      glBlendFunc (GL_SRC_ALPHA, GL_ONE);
    End;
  Texture.bind;
  DrawRectangle(X,Y,Texture.Width,Texture.Height);
  If Transparent then
    Begin
      glDisable(GL_BLEND);
//      glColor4f( 1.0, 1.0, 1.0, 1.0);
//      glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    End;
End;

(**
  Draws the Surface onto the Screen. Has been overridden to support multiple
  Different parameter collections. @ This implementation of draw enables the
  display of a section of the image at position X,Y of the base S2DLDraw
  surface. The section of the image to display is defined as the pattern
  number. The number of patterns in an image is defined by the PatternHeight
  and PatternWidth properties.

  @param   X       as an Integer
  @param   Y       as an Integer
  @param   Pattern as an Integer

**)
Procedure TS2DLImage.Draw(X,Y : Integer; Pattern : Integer);
Var
  W : Integer;
  FX,FY : Single;
Begin
  W := Texture.Width div PatternWidth;
  FX := (Pattern Mod W) * PatternWidthFrac;
  FY := (Pattern Div W) * PatternHeightFrac;
  If Transparent then
    Begin
      glEnable(GL_BLEND);
//      glColor4f( 1.0, 1.0, 1.0, 1.0);
//      glBlendFunc (GL_SRC_ALPHA, GL_ONE);
    End;
  Texture.bind;
  DrawRectangle(FX,FY,FX+PatternWidthFrac,FY+PatternHeightFrac,X,Y,PatternWidth,PatternHeight);
  If Transparent then
    Begin
      glDisable(GL_BLEND);
//      glColor4f( 1.0, 1.0, 1.0, 1.0);
//      glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    End;
End;

(**
  Draws the image tiled. Use the W and H to indicate the number of tiles
  to do as Width and Height.

  @param   X as an Integer
  @param   Y as an Integer
  @param   W as an Integer
  @param   H as an Integer
**)
procedure TS2DLImage.DrawTiled(X, Y, W, H: Integer);
Var
  OX, I, J : Integer;
begin
  OX := X;
  For I := 0 to H-1 do
    Begin
      X := OX;
      For J := 0 to W-1 do
        Begin
          Draw(X,Y);
          X := X + Texture.Width;
        End;
      Y := Y + Texture.Height;
    End;
end;

(**
  Draws a portion of image tiled. Use the W and H to indicate the number of tiles
  to do as Width and Height.

  @param   X as an Integer
  @param   Y as an Integer
  @param   W as an Integer
  @param   H as an Integer
  @param Pattern as an Integer
**)
procedure TS2DLImage.DrawTiled(X, Y, W, H, Pattern: Integer);
Var
  OX, I, J : Integer;
begin
  OX := X;
  For I := 0 to H-1 do
    Begin
      X := OX;
      For J := 0 to W-1 do
        Begin
          Draw(X,Y, Pattern);
          X := X + PatternWidth;
        End;
      Y := Y + PatternHeight;
    End;
end;

Type
  (**
    TPixelArray is Used in CopyBitmap to enable a Delphi TBitmap to be copied
    pixel by pixel onto the surface of an S2DLImage.
  **)
  TPixelArray = Array[0..64000] of UInt8;
  (**
    TPixelFormat is Used in CopyBitmap to enable a Delphi TBitmap to be copied
    pixel by pixel onto the surface of an S2DLImage.
  **)
  TPixelFormat = Array[0..2] of UInt8;

(**

  NOT USED

  @param   X     as an Integer
  @param   Y     as an Integer
  @param   Color as an UInt32

**)
procedure TS2DLImage.SetPixel(X, Y: Integer; Color: UInt32);
var
  bits : Pointer;
  bpp, r, g, b : Uint8;
begin
  If ImageSurface = Nil then
    Exit;
  if ( SDL_LockSurface( ImageSurface ) < 0 ) then
    exit;
  bpp := ImageSurface.format.BytesPerPixel;
  bits := @(TPixelArray(ImageSurface.Pixels^)[(Y * ImageSurface.pitch) + (X * bpp)]);
  // Set the pixel
  case bpp of
      1:
      begin
          PUint8( bits )^ := Uint8(Color);
      end;
      2:
      begin
          PUint16(bits)^ := Uint16(Color);
      end;
      3:
      begin // Format/endian independent
        r := (Color shr ImageSurface.format.Rshift ) and $FF;
        g := (Color shr ImageSurface.format.Gshift ) and $FF;
        b := (Color shr ImageSurface.format.Bshift ) and $FF;
        TPixelFormat(bits^)[(ImageSurface.format.Rshift div 8)] := r;
        TPixelFormat(bits^)[(ImageSurface.format.Gshift div 8)] := g;
        TPixelFormat(bits^)[(ImageSurface.format.Bshift div 8)] := b;
      end;
      4:
      begin
          PUInt32(bits)^ := Uint32( Color );
      end;
  end;
  // Update the display
  SDL_UnlockSurface(ImageSurface);
  SDL_UpdateRect(ImageSurface, X, Y, 1, 1 );
end;

(**
  Enables the drawing of a windows TBitmap object onto the Image Surface. This enables
  the drawing and manipulation of images using the standard windows canvas
  encapulated in a TBitmap to be done and then drawn onto the Images surface.

  @param   X                as an Integer
  @param   Y                as an Integer
  @param   BMP              as a TBitmap
  @param   TransparentColor as an UInt32
**)
{procedure TS2DLImage.CopyBitmap(X, Y: Integer; BMP: TBitmap;
  TransparentColor: UInt32);
var
  bits : Pointer;
  I,J, bpp, r, g, b : Uint8;
  Color : UInt32;
begin
  If ImageSurface = Nil then
    Exit;
  if ( SDL_LockSurface( ImageSurface ) < 0 ) then
    exit;
  bpp := ImageSurface.format.BytesPerPixel;
  For I := 0 to BMP.Width-1 do
    For J := 0 to BMP.Height-1 do
      Begin
        Color := BMP.Canvas.Pixels[I,J];
        If Color <> TransparentColor then
          Begin
            bits := @(TPixelArray(ImageSurface.Pixels^)[((Y+J) * ImageSurface.pitch) + ((X+I) * bpp)]);
            // Set the pixel
            case bpp of
              1:
              begin
                  PUint8( bits )^ := Uint8(Color);
              end;
              2:
              begin
                  PUint16(bits)^ := Uint16(Color);
              end;
              3:
             begin // Format/endian independent
                r := (Color shr ImageSurface.format.Rshift ) and $FF;
                g := (Color shr ImageSurface.format.Gshift ) and $FF;
                b := (Color shr ImageSurface.format.Bshift ) and $FF;
                TPixelFormat(bits^)[(ImageSurface.format.Rshift div 8)] := r;
                TPixelFormat(bits^)[(ImageSurface.format.Gshift div 8)] := g;
                TPixelFormat(bits^)[(ImageSurface.format.Bshift div 8)] := b;
              end;
              4:
              begin
                  PUInt32(bits)^ := Uint32( Color );
              end;
            end;
          End;
  End;
  // Update the display
  SDL_UnlockSurface(ImageSurface);
  SDL_UpdateRect(ImageSurface, X, Y, BMP.Width, BMP.Height);
end;}

(**
  This is a setter method for the PatternHeight property. This defines the tiles
  within an image.

  @param   Value as an Integer constant
**)
procedure TS2DLImage.SetPatternHeight(const Value: Integer);
begin
  FPatternHeight := Value;
  PatternHeightFrac := (FPatternHeight / Texture.Height);
  If PatternHeightFrac = 0 then
    PatternHeightFrac := 1;
end;

(**
  This is a setter method for the PatternWidth property. This defines the tiles
  within an image.

  @param   Value as an Integer constant
**)
procedure TS2DLImage.SetPatternWidth(const Value: Integer);
begin
  FPatternWidth := Value;
  PatternWidthFrac := (FPatternWidth / Texture.Width);
  If PatternWidthFrac = 0 then
    PatternWidthFrac := 1;
end;

(**
   Bind the image as the Current Texture.
**)
procedure TS2DLImage.BindImage;
begin
    If Assigned(Texture) then
      FreeAndNil(Texture); 
  if ( SDL_LockSurface( ImageSurface ) < 0 ) then
    exit;
  Texture := TTexture.Create;
  glGenTextures( 1, @Texture.Texture );
  // Typical Texture Generation Using Data From The Bitmap
  glBindTexture( GL_TEXTURE_2D, Texture.texture );

  // Generate The Texture
  if ImageSurface.format.bitsperpixel=24 then
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, ImageSurface.w, ImageSurface.h, 0, GL_RGB, GL_UNSIGNED_BYTE, ImageSurface.pixels )
  else
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, ImageSurface.w, ImageSurface.h, 0, GL_RGBA, GL_UNSIGNED_BYTE, ImageSurface.pixels );

  // Linear Filtering
  // scale linearly when image bigger than texture
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  // scale linearly when image smaller than texture
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  Texture.Width := ImageSurface.w;
  Texture.Height := ImageSurface.h;
  SDL_UnlockSurface(ImageSurface);
end;

(**
  Loads an image. The filename must be either absolute of based on current
  EXE location.

  @param   FileName as a String
**)
procedure TS2DLImage.LoadFromFile(FileName: String);
begin
  LoadFromFile(PChar(FileName));
end;

(**
  Draw the image to an area of the screen.

  @param   X  as an Integer
  @param   Y  as an Integer
  @param   IX as an Integer
  @param   IY as an Integer
  @param   IW as an Integer
  @param   IH as an Integer
**)
procedure TS2DLImage.DrawArea(X, Y, IX, IY, IW, IH: Integer);
Begin
  If Transparent then
    Begin
      glEnable(GL_BLEND);
      glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    End;
  Texture.bind;
  DrawRectangle(IX,IY,IX+IW,IY+IH,X,Y,IW,IH);
  If Transparent then
    Begin
      glDisable(GL_BLEND);
      glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    End;
end;

(**
  Loads an image. The filename must be either absolute of based on current
  EXE location.

  @param   FileName as a String
**)
procedure TS2DLImageList.LoadImage(FileName: String);
Begin
  LoadFromFile(FileName, FileName);
end;

(**
  Stretch the Image to the W and H defined.

  @param   X as an Integer
  @param   Y as an Integer
  @param   W as an Integer
  @param   H as an Integer
**)
procedure TS2DLImage.DrawStretch(X, Y, W, H: Integer);
begin
//  Draw(S2DLDraw,X,Y);
  If Transparent then
    glEnable(GL_BLEND);
  Texture.bind;
  DrawRectangle(X,Y,W,H);
  If Transparent then
    glDisable(GL_BLEND);
end;

(**
  Stretch a portion of the Image to the W and H defined.

  @param   X       as an Integer
  @param   Y       as an Integer
  @param   W       as an Integer
  @param   H       as an Integer
  @param   Pattern as an Integer
**)
procedure TS2DLImage.DrawStretch(X, Y, W, H, Pattern: Integer);
Var
  W2 : Integer;
  FX,FY : Single;
Begin
//  Draw(S2DLDraw,X,Y,Pattern);
  W2 := Texture.Width div PatternWidth;
  FX := (Pattern Mod W2) * PatternWidthFrac;
  FY := (Pattern Div W2) * PatternHeightFrac;
  If Transparent then
    glEnable(GL_BLEND);
  Texture.bind;
  DrawRectangle(FX,FY,FX+PatternWidthFrac,FY+PatternHeightFrac,X,Y,W,H);
  If Transparent then
    glDisable(GL_BLEND);
end;

(**
  Draw a rectangular area from the image on the screen

  @param   X       as an Integer
  @param   Y       as an Integer
  @param   W       as an Integer
  @param   H       as an Integer
  @param   X1      as an Single
  @param   Y1      as an Single
  @param   X2      as an Single
  @param   Y1      as an Single
**)
procedure TS2DLImage.DrawRect(X, Y, W, H: Integer; X1, Y1, X2, Y2: Single);

Begin
  If Transparent then
    glEnable(GL_BLEND);
  Texture.bind;
  DrawRectangle(X1,Y1,X2,Y2,X,Y,W,H);
  If Transparent then
    glDisable(GL_BLEND);
end;

(**
  Draw the Image rotated by a number of degrees.

  @param   X     as an Integer
  @param   Y     as an Integer
  @param   Angle as a Single
**)
procedure TS2DLImage.DrawRotate(X, Y: Integer; Angle: Single);
begin
  If Transparent then
    Begin
      glEnable(GL_BLEND);
    End;
  Texture.bind;
  DrawRectangle(X,Y,Texture.Width,Texture.Height,Angle);
  If Transparent then
    Begin
      glDisable(GL_BLEND);
    End;
end;

(**
    Draw a portion of the Image rotated by a number of degrees.

  @param   X       as an Integer
  @param   Y       as an Integer
  @param   Pattern as an Integer
  @param   Angle   as a Single
**)
procedure TS2DLImage.DrawRotate(X, Y, Pattern: Integer; Angle: Single);
Var
  W : Integer;
  FX,FY : Single;
Begin
  W := Texture.Width div PatternWidth;
  FX := (Pattern Mod W) * PatternWidthFrac;
  FY := (Pattern Div W) * PatternHeightFrac;
  If Transparent then
    Begin
      glEnable(GL_BLEND);
    End;
  Texture.bind;
  DrawRectangle(FX,FY,FX+PatternWidthFrac,FY+PatternHeightFrac,X,Y,PatternWidth,PatternHeight, Angle);
  If Transparent then
    Begin
      glDisable(GL_BLEND);
    End;
end;

(**
  Draw the Image using alpha Blending.

  @param   X     as an Integer
  @param   Y     as an Integer
  @param   Alpha as a Single
**)
procedure TS2DLImage.DrawAlpha(X, Y: Integer; Alpha: Single);
begin
  glEnable(GL_BLEND);
  glColor4f( 1.0, 1.0, 1.0, Alpha);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  Draw(X,Y);
//  Texture.bind;
//  DrawRectangle(X,Y,Texture.Width,Texture.Height);
  glDisable(GL_BLEND);
  glColor4f( 1.0, 1.0, 1.0, 1.0);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
end;

(**
    Draw the portion of the Image using alpha Blending.

  @param   X       as an Integer
  @param   Y       as an Integer
  @param   Pattern as an Integer
  @param   Alpha   as a Single
**)
procedure TS2DLImage.DrawAlpha(X, Y, Pattern: Integer; Alpha: Single);
Var
  W : Integer;
  FX,FY : Single;
Begin
  W := Texture.Width div PatternWidth;
  FX := (Pattern Mod W) * PatternWidthFrac;
  FY := (Pattern Div W) * PatternHeightFrac;
  glEnable(GL_BLEND);
  glColor4f( 1.0, 1.0, 1.0, Alpha);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  Texture.bind;
  DrawRectangle(FX,FY,FX+PatternWidthFrac,FY+PatternHeightFrac,X,Y,PatternWidth,PatternHeight);
  glDisable(GL_BLEND);
  glColor4f( 1.0, 1.0, 1.0, 1.0);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
end;

function TS2DLImageList.DefaultGetAnImage(Index: Variant): TS2DLImage;
begin
  If varType(Index) = varString then
    Begin
      Result := ItemByName[Index];
    End
  Else
    Begin
      Result := Items[Index];
    End;
end;

procedure TS2DLImage.LoadFromFile(FileName: String; IsTransparent: Boolean);
begin
  If FileExists(FileName) then
    Begin
      Texture := TTexture.Create;
      Texture.loadFrom(FileName);
      Transparent := IsTransparent;
    End
  Else
    Begin
      Log.LogError('File:'+FileName+' not found','TS2DLImage.LoadFromFile');
    End;
end;

(**
  Load an image into memory and return the Image.

  @param   Name     as a String
  @param   FileName as a String
  @return  a TS2DLImage
**)
function TS2DLImageList.LoadFromFile(FileName, Name: String): TS2DLImage;
Var
  Image : TS2DLImage;
Begin
  // Start Loading Images
  Log.LogStatus('Loading:'+Name,'ImageList.LoadImage');
  Image := TS2DLImage.Create;
  Image.LoadFromFile(PChar(FileName));
  Image.Transparent := True;
  Image.Name := Name;
  Add(Image);
  Result := Image;
end;

procedure TS2DLImage.DrawStretchRotate(X, Y, W, H, Pattern, Degrees: Integer);
Var
  W2 : Integer;
  FX,FY : Single;
Begin
//  Draw(S2DLDraw,X,Y,Pattern);
  W2 := Texture.Width div PatternWidth;
  FX := (Pattern Mod W2) * PatternWidthFrac;
  FY := (Pattern Div W2) * PatternHeightFrac;
  If Transparent then
    glEnable(GL_BLEND);
  Texture.bind;
  DrawRectangle(FX,FY,FX+PatternWidthFrac,FY+PatternHeightFrac,X,Y,W,H,Degrees);
  If Transparent then
    glDisable(GL_BLEND);
end;

procedure TS2DLImage.DrawStretchAlpha(X, Y, W, H, Pattern: Integer;
  Alpha: Single);
Var
  W2 : Integer;
  FX,FY : Single;
Begin
//  Draw(S2DLDraw,X,Y,Pattern);
  W2 := Texture.Width div PatternWidth;
  FX := (Pattern Mod W2) * PatternWidthFrac;
  FY := (Pattern Div W2) * PatternHeightFrac;
{  If Transparent then
    glEnable(GL_BLEND);}
  glEnable(GL_BLEND);
  glColor4f( 1.0, 1.0, 1.0, Alpha);
{  glColor4b( 127, 127, 127, -128+round(Alpha*255));}
  glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  Texture.bind;
  DrawRectangle(FX,FY,FX+PatternWidthFrac,FY+PatternHeightFrac,X,Y,W,H,Alpha);
{  If Transparent then
    glDisable(GL_BLEND);}
  glColor4f( 1.0, 1.0, 1.0, 1.0);
{  glColor4b( 127, 127, 127, 127);}
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
end;

procedure TS2DLImage.DrawStretchRGBA(X, Y, W, H:integer;R,G,B,A:single);

Begin
{  If Transparent then
    glEnable(GL_BLEND);}
  glEnable(GL_BLEND);
  glColor4f(R,G,B,A);
{  glColor4b( 127, 127, 127, -128+round(Alpha*255));}
  glBlendFunc (GL_SRC_ALPHA, GL_ONE);
  Texture.bind;
  DrawRectangle(X,Y,W,H);
{  DrawRectangle(X,Y,W,H,X,Y,W,H,Alpha);}
{  If Transparent then
    glDisable(GL_BLEND);}
  glColor4f( 1.0, 1.0, 1.0, 1.0);
{  glColor4b( 127, 127, 127, 127);}
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
end;

Initialization
  Images := TS2DLImageList.Create;

Finalization
  Images.Free;

End.
