(** S2DLNeheFont
@Name: Simple 2D Library - Bitmapped Fonts Unit
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.11
@Change_History:
@  ver 1.11 - Working on Linux (Daniel Mantione)
@  ver 1.11 - Added WriteTextStretch
@  ver 1.10 - First release version

@Copyright: (c) 2004-2007 by William Cairns (CairnsGames)

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGD web site.

@Usage: Create a PNG file with Alpha channel supporting the limited character
@ set as suggested by the Nehe pages. (See included example files). 
@Note: Do not initialize Fonts:TS2DLNeheFontList yourself !!! - Automatically done !!!

@Author: W.Cairns **)
unit S2DLNeheFont;

{$ifdef FPC}
{$mode delphi}
{$endif}

interface

Uses
  Classes, S2DLImage, IniFiles, SysUtils, Variants;

Type
  (** Stores an individual font file (can include 2 font sets) **)
  TS2DLNeheFont = Class;
  (** Manages a list of available fonts **)
  TS2DLNeheFontList = Class
    Private
      FItems : TList;
      Function GetFont(Index : Integer) : TS2DLNeheFont;
      Function GetFontByName(Name : String) : TS2DLNeheFont;
      function GetGetAFont(Index: Variant): TS2DLNeheFont;
    Public
      Property Items[Index : Integer] : TS2DLNeheFont read GetFont;
      Property ItemByName[Name : String] : TS2DLNeheFont read GetFontByName;
      Property GetAFont[Index : Variant] : TS2DLNeheFont read GetGetAFont; default;
      Constructor Create;
      Destructor Destroy; Override;
      Function Find(Name : String) : TS2DLNeheFont;
      Procedure Add(Item : TS2DLNeheFont);
      Function Count : Integer;
      Function Delete(Item : Integer) : Boolean;
      Procedure LoadFont(FontName : String; FontSize : Integer = 24; Bold : Boolean = False); Overload;
      Procedure LoadFont(FileName : String; FontName : String); Overload;
      Function LoadFromFile(FileName,FontName : String) : TS2DLNeheFont;
  End;
  (** Stores an individual font file (can include 2 font sets) **)
  TS2DLNeheFont = Class
  private
    FFontName: String;
    CharWidths : Array[0..255] of Byte;
    CharHeights : Array[0..255] of Byte;
    CharX1 : Array[0..255] of Single;
    CharY1 : Array[0..255] of Single;
    CharX2 : Array[0..255] of Single;
    CharY2 : Array[0..255] of Single;
    procedure SetFontName(const Value: String);
  public
    Image : TS2DLImage;
    Constructor Create;
    Procedure WriteText(X,Y : Integer; Text : String; Mode : Integer = 0);
    procedure WriteTextStretch(X, Y: Integer; Ratio: Single; Text: String; Mode: Integer = 0);
    Procedure WriteAlpha(X,Y : Integer; Text : String; Alpha : Single; Mode : Integer = 0);
    Procedure RenderFont(X,Y : Integer; Text : String; Mode : Integer = 0);
    Procedure Write(X,Y : Integer; Text : String; Mode : Integer = 0);
    Property FontName : String read FFontName write SetFontName;
    Procedure LoadIni(FileName : String);
  Public
    FontWidth1 : Single;
  End;

Var
  Fonts : TS2DLNeheFontList;

implementation

{ TS2DLNeheFontList }

(**
  This is the constructor method for the TS2DLNeheFontList class.
  Allocates memory to store lists. (Using a list structure).
**)
Constructor TS2DLNeheFontList.Create;
Begin
  FItems := TList.Create;
End;

(**
  This is the destructor method for the TS2DLNeheFontList class.
**)
Destructor TS2DLNeheFontList.Destroy;
Var
  I : Integer;
Begin
  // Free memory used for each image
  For I := FItems.Count -1 downto 0 do
    Begin
      TS2DLNeheFont(FItems[I]).Free;
      FItems.Delete(I);
    End;
  FItems.Free;
End;

(**
  Find the Font identified by Index.
  @see Find
  @param   Index as an Integer
  @return  a TS2DLNeheFont
**)
Function TS2DLNeheFontList.GetFont(Index : Integer) : TS2DLNeheFont;
Begin
  Result := TS2DLNeheFont(FItems[Index]);
End;

(**
  This is a getter method for the FontByName property.
  @param   Name as a String
  @return  a TS2DLNeheFont
**)
Function TS2DLNeheFontList.GetFontByName(Name : String) : TS2DLNeheFont;
Var
  I : Integer;
Begin
  For I := FItems.Count-1 downto 0 do
    Begin
      If TS2DLNeheFont(FItems[I]).FontName = Name then
        Begin
          Result := TS2DLNeheFont(FItems[I]);
          Exit;
        End;
    End;
End;

(**
  Find the Font identified by Name. Return the Font object not an index value.
  This allows the use of Font.Find(FontName).<method>

  @param   Name as a String
  @return  a TS2DLNeheFont
**)
Function TS2DLNeheFontList.Find(Name : String) : TS2DLNeheFont;
Begin
  Result := GetFontByName(Name);
End;

(**
  Add a new font to the Font list.
  @param   Item as a TS2DLNeheFont
**)
Procedure TS2DLNeheFontList.Add(Item : TS2DLNeheFont);
Begin
  FItems.Add(Item);
End;

(**
  Returns the number of fonts available in the list.
  @return  an Integer
**)
Function TS2DLNeheFontList.Count : Integer;
Begin
  Result := FItems.Count;
End;

(**
  Remove a font from the list. This also frees the memory for the font that
  was stored in the list.
  @param   Item as an Integer
  @return  a Boolean
**)
Function TS2DLNeheFontList.Delete(Item : Integer) : Boolean;
Var
  S2DLFont : TS2DLNeheFont;
Begin
  S2DLFont := TS2DLNeheFont(FItems[Item]);
  FItems.Delete(Item);
  S2DLFont.Free;
  Result := True;
End;

(**
  Creates a new TS2DLNeheFont object and loads it into the Font List.
  Then loads the Font specified by FontName into the newly created font.

  @param   FontName as a String
  @param   FontSize as an Integer
  @param   Bold     as a Boolean
**)
procedure TS2DLNeheFontList.LoadFont(FontName: String; FontSize: Integer;
  Bold: Boolean);
Var
  Font : TS2DLNeheFont;
begin
  Font := TS2DLNeheFont.Create;
  Font.FontName := PChar(FontName);
  Font.Image := TS2DLImage.Create;
  Font.Image.LoadFromFile(FontName);
  Font.Image.Transparent := True;
  Font.Image.PatternWidth := Font.Image.Width div 16;
  Font.Image.PatternHeight := Font.Image.Height div 16;
  Fonts.Add(Font);
end;

{ TS2DLNeheFont }

(**
  This is the constructor method for the TS2DLNeheFont class.
**)
constructor TS2DLNeheFont.Create;
Var
  I : Integer;
begin
  FontWidth1 := 0.60;
  For I := 0 to 255 do
    begin
      CharWidths[I] := 0;
      CharHeights[I] := 0;
      CharX1[I] := -1;
      CharY1[I] := -1;
      CharX2[I] := -1;
      CharY2[I] := -1;
    end;
end;

(**
  Loads the images sizes from an INI file. Note that the default settings from
  the example Bitmap Font builder are not always correct.

  @param   FileName as a String
**)
procedure TS2DLNeheFont.LoadIni(FileName: String);
Var
  Ini : TIniFile;
  I,n,c : Integer;
begin
  Ini := TIniFile.Create(FileName);
  If Ini.SectionExists('Char Widths') Then
    begin
      For I := 0 to 255 do
        CharWidths[I] := Ini.ReadInteger('Char Widths',IntToStr(I),0);
    end
  Else
    begin
      n:=0;
      repeat
        c:=Ini.ReadInteger('FontData',IntToStr(n)+'Char',-1);
        If c=-1 Then
          break;
        CharHeights[c] := Ini.ReadInteger('FontData',IntToStr(n)+'Hgt',0);
        CharWidths[c] := Ini.ReadInteger('FontData',IntToStr(n)+'Wid',0);
        CharX1[c] := Ini.ReadFloat('FontData',IntToStr(n)+'X1',0);
        CharY1[c] := Ini.ReadFloat('FontData',IntToStr(n)+'Y1',0);
        CharX2[c] := Ini.ReadFloat('FontData',IntToStr(n)+'X2',0);
        CharY2[c] := Ini.ReadFloat('FontData',IntToStr(n)+'Y2',0);
        inc(n);
      until false;
    end;
  Ini.Free;
end;

(**
  Draw the text onto the screen.

  @param   X    as an Integer
  @param   Y    as an Integer
  @param   Text as a String
  @param   Mode as an Integer
**)
procedure TS2DLNeheFont.RenderFont(X, Y: Integer; Text: String;
  Mode: Integer);
begin
  WriteText(X,Y,Text,Mode);
end;

(**
  This is a setter method for the FontName property.

  @param   Value as a String constant
**)
procedure TS2DLNeheFont.SetFontName(const Value: String);
begin
  FFontName := Value;
end;

(**
    Draw the text onto the screen.

  @param   X    as an Integer
  @param   Y    as an Integer
  @param   Text as a String
  @param   Mode as an Integer
**)
procedure TS2DLNeheFont.Write(X, Y: Integer; Text: String; Mode: Integer);
begin
  WriteText(X,Y,Text,Mode);
end;

(**
  Draw the text onto the screen.

  @param   X    as an Integer
  @param   Y    as an Integer
  @param   Text as a String
  @param   Mode as an Integer
**)
procedure TS2DLNeheFont.WriteAlpha(X, Y: Integer; Text: String;
  Alpha: Single; Mode: Integer);
Var
  I,P : Integer;
  C : Integer;
begin
  For I := 1 to Length(Text) do
    Begin
      C := Ord(Text[I])-32;
      If Mode = 1 then
        C := C + 128;
      Image.DrawAlpha(X,Y,C,Alpha);
      If CharWidths[C] = 0 then
        Begin
          If Mode = 0 then
            X := X + Trunc(Image.Width * FontWidth1)
          Else
            X := X + Trunc(Image.Width * 0.39);
        End
      Else
        X := X + CharWidths[C];
    End;
end;

procedure TS2DLNeheFont.WriteText(X, Y: Integer; Text: String;
  Mode: Integer);
Var
  I,P : Integer;
  C : Integer;
begin
  For I := 1 to Length(Text) do
    Begin
      C := Ord(Text[I])-32;
      If Mode = 1 then
        C := C + 128;
      Image.Draw(X,Y,C);
      If CharWidths[C] = 0 then
        Begin
          If Mode = 0 then
            X := X + Trunc(Image.Width * FontWidth1)
          Else
            X := X + Trunc(Image.Width * 0.39);
        End
      Else
        X := X + CharWidths[C];
    End;
end;

procedure TS2DLNeheFont.WriteTextStretch(X, Y: Integer; Ratio : Single; Text: String;
  Mode: Integer);
Var
  I,P : Integer;
  C : Integer;
  TW : Integer;
begin
  If Mode = 0 then
    TW := Trunc((Image.Width * FontWidth1) * Ratio)
  Else
    TW := Trunc((Image.Width * 0.39) * Ratio);
  For I := 1 to Length(Text) do
    Begin
      c:=Byte(Text[I]);
      If CharX1[c]>=0 then
        begin
          Image.DrawRect(X,Y,trunc(ratio*CharWidths[C]),trunc(ratio*CharHeights[C]),
                         CharX1[C],CharY1[C],CharX2[C],CharY2[C]);
        end
      else
        begin
          C := Ord(Text[I])-32;
          If Mode = 1 then
            C := C + 128;
          Image.DrawStretch(X,Y,TW,TW,C);
        end;
      If CharWidths[C] = 0 then
        Begin
          X := X + TW;
        End
      Else
        X := X + Trunc(CharWidths[C] * Ratio);
    End;
end;

(**
  Load a new font into memory. The font can then be access by Name or
  by index.

  @param   FileName as a String
  @param   FontName as a String
**)
procedure TS2DLNeheFontList.LoadFont(FileName, FontName: String);
Var
  Font : TS2DLNeheFont;
begin
  Font := TS2DLNeheFont.Create;
  Font.FontName := PChar(FontName);
  Font.Image := TS2DLImage.Create;
  Font.Image.LoadFromFile(FileName);
  Font.Image.Transparent := True;
  Font.Image.PatternWidth := Font.Image.Width div 16;
  Font.Image.PatternHeight := Font.Image.Height div 16;
  Fonts.Add(Font);
end;

(**
  Internal method to allow access of fonts by Name or by Index
  while using the same default property. This is ineffect a passthrough method.

  @param   Index as a Variant
  @return  a TS2DLNeheFont
**)
function TS2DLNeheFontList.GetGetAFont(Index: Variant): TS2DLNeheFont;
begin
  If varType(Index) = varString then
    Begin
      Result := ItemByName[Index];
    End
  Else
    Begin
      Result := Items[Index];
    End;
end;

(**
  Load a Font into memory and return the new font.

  @param   FontName as a String
  @param   FileName as a String
  @return  a TS2DLNeheFont
**)
function TS2DLNeheFontList.LoadFromFile(FileName, FontName: String): TS2DLNeheFont;
Var
  Font : TS2DLNeheFont;
begin
  Font := TS2DLNeheFont.Create;
  Font.FontName := PChar(FontName);
  Font.Image := TS2DLImage.Create;
  Font.Image.LoadFromFile(FileName);
  Font.Image.Transparent := True;
  Font.Image.PatternWidth := Font.Image.Width div 16;
  Font.Image.PatternHeight := Font.Image.Height div 16;
  Fonts.Add(Font);
  Result := Font;
end;

Initialization
  Fonts := TS2DLNeheFontList.Create;

Finalization
  Fonts.Free;

end.
