(** S2DLDraws
@Name: Simple 2D Library - Sounds
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.11

@Change_History: 
@  ver 1.11 - Working on Linux (Daniel Mantione)
@  ver 1.10 - support to disable sound based on settings.
@Copyright: (c) 2004-2007 by William Cairns (CairnsGames)

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGD web site.

@Usage: Automatically creates a sound engine. You need to load sounds into the
  sound engine and then call the sounds play method.

@Author: W.Cairns **)
unit S2DLSound;

{$ifdef FPC}
{$mode delphi}
{$endif}

interface

uses
  Classes, SysUtils, S2DLLogger, sdl, sdl_mixer,
  Windows;
  
Type
  (**
    The Sound is stored in this object which supplies methods to play the sound.
  **)
  TS2DLSound = Class;
  (**
    Sound Engine stores a list of sounds and has methods and properties to make
    it easy to access and use the sounds.
  **)
  TS2DLSoundEngine = Class
    Private
      FItems : TList;
      FVolume : Integer;
      Function GetSound(Index : Integer) : TS2DLSound;
      Function GetSoundByName(Name : String) : TS2DLSound;
      procedure SetVolume(const AValue : Integer);
      function DefaultGetASound(Index: Variant): TS2DLSound;
    Public
      Property Sound[Index : Integer] : TS2DLSound read GetSound;
      Property SoundByName[Name : String] : TS2DLSound read GetSoundByName;
      Property Items[Index : Integer] : TS2DLSound read GetSound;
      Property ItemByName[Name : String] : TS2DLSound read GetSoundByName;
      Property GetASound[Index : Variant] : TS2DLSound read DefaultGetASound; Default;
      Property Volume : Integer read FVolume write SetVolume;
      Constructor Create;
      Destructor Destroy; Override;
      Function Find(Name : String) : TS2DLSound;
      Procedure FindAndPlay(FileName : String; Name : String);
      Procedure Add(Item : TS2DLSound);
      Function Count : Integer;
      Function Delete(Item : Integer) : Boolean;
      Procedure LoadSound(Name : String; FileName : String); Overload;
      Function LoadFromFile(FileName : String; Name : String) : TS2DLSound; Overload;
  End;
  (**
    The Sound is stored in this object which supplies methods to play the sound.
  **)
  TS2DLSound = Class
    Name : String;
    Sound: PMix_Chunk;
    Volume : Integer;
    Engine : TS2DLSoundEngine;
    Channel : Integer;
    LastPlay : Cardinal;
    Constructor Create;
    Destructor Destroy; Override;
    Procedure LoadFromFile(FileName : PChar);
    Procedure Play(NewVolume : Integer = 100); Overload;
    Procedure Play(NewChannel : Integer; NewVolume : Integer); Overload;
    Procedure Stop;
  End;

Var
  Sounds : TS2DLSoundEngine;

procedure PlaySound(Name: String);
  
  
implementation

{ TS2DLSound }

Uses
  Variants;

Var
  CanSound : Boolean;

(**
  This is the constructor method for the TS2DLSound class.
**)
constructor TS2DLSound.Create;
begin

end;

(**
  This is the destructor method for the TS2DLSound class.
  Frees SDL sound information.
**)
destructor TS2DLSound.Destroy;
begin
  Mix_FreeChunk(Sound);

  inherited Destroy;
end;

(**
  Load the WAV file identified by FileName.
  Does not set the Sound's Name property.
  @param   FileName as a PChar
**)
procedure TS2DLSound.LoadFromFile(FileName : PChar);
begin
  Sound := Mix_LoadWAV(PChar(FileName));
  If Sound = Nil then
    Begin
      Sound := Nil;
      Log.LogError('Sound '+FileNAme+' not found.','TS2DLSound.LoadFromFile');
    End;
end;

(**
  Play a sound at a new volume
  @param   NewVolume as an Integer
**)
procedure TS2DLSound.Play(NewVolume : Integer = 100);
begin
  If GetTickCount - LastPlay < 500 then
    Exit;
  LastPlay := GetTickCount;
  Channel := Mix_PlayChannel(-1, Sound, 0);
  Mix_Volume(Channel, NewVolume);
end;

(**
  Play the sound on a specific channel and volume.
  @param   NewChannel as an Integer
  @param   NewVolume  as an Integer
**)
procedure TS2DLSound.Play(NewChannel : Integer; NewVolume : Integer);
begin
  If GetTickCount - LastPlay < 500 then
    Exit;
  LastPlay := GetTickCount;
  Channel := Mix_PlayChannel(NewChannel, Sound, 0);
  Mix_Volume(Channel, NewVolume);
end;

(**
  Stop the Sound.
**)
procedure TS2DLSound.Stop;
begin
  Mix_HaltChannel(Channel);
end;

{ TS2DLSoundEngine }

(**
  This is a getter method for the Sound property.
  Returns the sound identified by Index.
  @param   Index as an Integer
  @return  a TS2DLSound
**)
function TS2DLSoundEngine.GetSound(Index : Integer) : TS2DLSound;
begin
  If Not CanSound then
    Result := Nil
  Else
  Result := TS2DLSound(FItems[Index]);
end;

(**
  This is a getter method for the SoundByName property.
  @param   Name as a String
  @return  a TS2DLSound
**)
function TS2DLSoundEngine.GetSoundByName(Name : String) : TS2DLSound;
Var
  I : Integer;
begin
  If CanSound then
    Begin
      For I := 0 to FItems.Count - 1 do
        Begin
          If TS2DLSound(FItems[I]).Name = Name then
            Begin
              Result := TS2DLSound(FItems[I]);
              Exit;
            End;
        end;
    End;
  Result := Nil;
end;

(**
  This is a setter method for the Volume property.
  @param   AValue as an Integer constant
**)
procedure TS2DLSoundEngine.SetVolume(const AValue : Integer);
begin
  if FVolume = AValue then
    exit;
  FVolume := AValue;
end;

(**
  This is the constructor method for the TS2DLSoundEngine class.
  Allocates memory for the Sound List.
**)
constructor TS2DLSoundEngine.Create;
begin
  FItems := TList.Create;
end;

(**
  This is the destructor method for the TS2DLSoundEngine class.
  For all sounds in the list free and remove from the list.
  Ensures that all memory used by the list and sound sin the list is freed.
**)
destructor TS2DLSoundEngine.Destroy;
begin
  While FItems.Count > 0 do
    Begin
      TS2DLSound(FItems[0]).Free;
      FItems.Delete(0);
    End;
  inherited Destroy;
end;

(**
  Returns the Sound object for the item with the specified name. Returns the
  sound object not the index allowing calls such as Sounds.Find(<Name>).Play
  @param   Name as a String
  @return  a TS2DLSound
**)
function TS2DLSoundEngine.Find(Name : String) : TS2DLSound;
begin
  If Not CanSound then
    Result := Nil
  Else
    Result := GetsoundByName(Name);
end;

(**
  Add a new sound item into the sound list. Sets the Sounds engine to the
  current list. 
  @param   Item as a TS2DLSound
**)
procedure TS2DLSoundEngine.Add(Item : TS2DLSound);
begin
  If Not CanSound then
    Exit;
  FItems.Add(Item);
  Item.Engine := Self;
end;

(**
  Returns the number of Sounds in the SoundList.
  @return  an Integer
**)
function TS2DLSoundEngine.Count : Integer;
begin
  Result := FItems.Count;
end;

(**
  Deletes the item from the list and frees its memory.
  @param   Item as an Integer
  @return  a Boolean
**)
function TS2DLSoundEngine.Delete(Item : Integer) : Boolean;
begin
  Result := False;
  try
    TS2DLSound(FItems[Item]).Free;
    FItems.Delete(Item);
    Result := True;
  except
    On E:Exception do
      Begin
        Log.LogWarning('Cannot delete sound with index:'+IntToStr(Item)+'.','TS2DLSoundEngine.Delete');
      End;
  end;
end;

(**
  Creates and Loads a sound into Memory.
  @param   Name     as a String
  @param   FileName as a String

**)
procedure TS2DLSoundEngine.LoadSound(Name: String; FileName: String);
Var
  Sound : TS2DLSound;
Begin
  try
    If Not CanSound then
      Exit;
    Sound := TS2DLSound.Create;
    Sound.LoadFromFile(PChar(FileName));
    Sound.Name := Name;
    Add(Sound);
  except
    On E:Exception do
      Begin
        Log.LogError('Sound '+FileName+' not found.','TS2DLSoundEngine.LoadSound');
      End;
  end;
end;

(**
  Play the sound using the a volume of 100. If the sound is not avaialble in the
  Sounds list it is loaded into memory and played.
  Note: If not in memory it is presumed to be a wav with the same name.
  @param   Name as a String
**)
procedure PlaySound(Name: String);
Var
  Sound : TS2DLSound;
begin
  Sound := Sounds.Find(Name);
  If Sound = Nil then
    Begin
      Sounds.LoadSound(Name,'sound'+directoryseparator+Name+'.wav');
      Sound := Sounds.Find(Name);
    End;
  Sound.Play(100);
end;

(**
  Load the Sound into Memory and allocate it a name. Return the created sound.

  @param   FileName as a String
  @param   Name     as a String
  @return  a TS2DLSound
**)
function TS2DLSoundEngine.LoadFromFile(FileName, Name: String): TS2DLSound;
Var
  Sound : TS2DLSound;
Begin
  Result := Nil;
  try
    If Not CanSound then
      Begin
        Result := Nil;
        Exit;
      End;
    Sound := TS2DLSound.Create;
    Sound.LoadFromFile(PChar(FileName));
    Sound.Name := Name;
    Add(Sound);
    Result := Sound;
  except
    On E:Exception do
      Begin
        Log.LogError('Sound '+FileName+' not found.','TS2DLSoundEngine.LoadFromFile');
      End;
  end;
end;

(**
  Find the sound in the Engine and play the sound immediatly.

  @param   FileName as a String
  @param   Name     as a String
**)
procedure TS2DLSoundEngine.FindAndPlay(FileName : String; Name: String);
Var
  S : TS2DLSound;
begin
  If CanSound then
    Begin
      S := Find(Name);
      If S = Nil then
        Begin
          S := LoadFromFile(FileName, Name);
        End;
      If S <> Nil then
        S.Play;
    End;
end;

function TS2DLSoundEngine.DefaultGetASound(Index: Variant): TS2DLSound;
begin
  If varType(Index) = varString then
    Begin
      Result := ItemByName[Index];
    End
  Else
    Begin
      Result := Items[Index];
    End;
end;

Initialization
  CanSound := SoundOn;
  if (Mix_OpenAudio(22050, AUDIO_U8, 1, 512) < 0) then
    begin
      Log.LogError(Format('Couldn''t set 22050 Hz 8-bit audio - Reason : %s',
        [Mix_GetError]), 'Main');
    end
  Else
    Sounds := TS2DLSoundEngine.Create;

Finalization
  Sounds.Free;
  Mix_CloseAudio();

end.

