(** S2DLState
@Name: Simple 2D Library - State Management System
  This file is part of the Free Pascal Simple 2D Libraries
@Version: 1.10
@Change_History:
@   ver 1.10 - Created Base State
@            - Create Static Image State
@            - Created Splash Screen State
@            - Implemented Keyboard and Mouse Deltas
@Copyright: (c) 2004-2007 by William Cairns (CairnsGames)

@License: This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
       (I got this from the JEDI-SDL files and it sounds Cool!)
@
By using this work you have immediatly agreed with the above.
@
 The work can be used for any purpose. You may sell it, use it in a game
 delete it or change it as you wish with no reward to the the original
 developers.
@
 We do however ask that if this work is used that some mention of it be made
 on the PGDev web site.

@Usage: Inherit a new state from the Base State. Implement the DoDraw, DoInput 
    and DoUpdate methods for the new state. Include the new state into the Game 
    Program. When a state ends it calls onEndState which must be implemented in 
    the main program to manage state transitions.

@Author: W.Cairns **)
unit S2DLState;

{$ifdef FPC}
{$mode objFPC}
{$endif}

{$h+}

interface

Uses
  SDL,
  SysUtils,
  Classes,
  S2DLImage,
  S2DLNeheFont,
  S2DLDraws;

Const
  rt_continue=0;
  rt_quit=1;

Type
  TMouse = Class
    X,Y : Integer;
    LeftButton, RightButton, LeftPressed, LeftReleased, RightPressed, RightReleased : Boolean;

  End;
  TKeyboard = Class
    KeyState : array[SDLK_FIRST..SDLK_EURO] of Boolean;
    KeyPressed : array[SDLK_FIRST..SDLK_EURO] of Boolean;
    KeyReleased : array[SDLK_FIRST..SDLK_EURO] of Boolean;
    Constructor Create;
    procedure reset;
  End;

  TGameState = class
  private
    FGameTitle: String;
    FCursor: TS2DLImage;
    FFont: TS2DLNeheFont;
    FActive: Boolean;
    FCursorOn: Boolean;
    FOnEndState: TNotifyEvent;
    FMouse: TMouse;
    FKeyboard: TKeyboard;
    FLastTick: Cardinal;
    procedure SetCursor(const Value: TS2DLImage);
    procedure SetFont(const Value: TS2DLNeheFont);
    procedure SetGameTitle(const Value: String);
    procedure SetActive(const Value: Boolean);
    procedure SetCursorOn(const Value: Boolean);
    procedure SetOnEndState(const Value: TNotifyEvent);
    procedure SetMouse(const Value: TMouse);
    procedure SetKeyboard(const Value: TKeyboard);
    procedure SetLastTick(const Value: Cardinal);
  public
    ReturnCode:integer;
    Property Font : TS2DLNeheFont read FFont write SetFont;
    Property GameTitle : String read FGameTitle write SetGameTitle;
    Property CursorOn : Boolean read FCursorOn write SetCursorOn;
    Property Active : Boolean read FActive write SetActive;
    Property OnEndState : TNotifyEvent read FOnEndState write SetOnEndState;
    Property Mouse : TMouse read FMouse write SetMouse;
    Property Keyboard : TKeyboard read FKeyboard write SetKeyboard;
    Property LastTick : Cardinal read FLastTick write SetLastTick;
    constructor Create;
    procedure DoDraw; Virtual;
    procedure DoEndState; Virtual;
    procedure DoInput; Virtual;
    procedure DoStartState; Virtual;
    procedure DoUpdate(GameTime : LongInt); Virtual;
    procedure Draw;
    procedure Initialise(NewImages: TS2DLImageList);
    procedure Input;
    procedure LoadResources;
    procedure Update;
    End;

  TGameStateImage = Class(TGameState)
  private
    FBackImage: TS2DLImage;
    procedure SetBackImage(const Value: TS2DLImage);
  public
    Property BackImage : TS2DLImage read FBackImage write SetBackImage;
    procedure DoDraw; Override;
    procedure DoInput; Override;
  End;

  TGameStateSplash = Class(TGameStateImage)
  private
    FDelay: LongInt;
    FDelayTime: LongInt;
    procedure DoDraw; Override;
    procedure DoEndState; Override;
    procedure DoInput; Override;
    procedure DoStartState; Override;
    procedure DoUpdate(GameTime: Integer); Override;
    procedure SetDelay(const Value: LongInt);
    procedure SetDelayTime(const Value: LongInt);
  public
    Constructor Create;
    Property Delay : LongInt read FDelay write SetDelay;
    Property DelayTime : LongInt read FDelayTime write SetDelayTime;
  End;

Const
  Pressed = True;

implementation

Constructor TGameState.Create;
Begin
    CursorOn := false;
    Keyboard := TKeyboard.Create;
    Mouse := TMouse.Create;
    ReturnCode := rt_continue;
End;

/// <summary>
/// Initialize all the required fields of the GameState.
/// This must be called after the game state is created.
/// </summary>
/// <param name="NewGraphics"></param>
/// <param name="NewContent"></param>
/// <param name="NewSpriteBatch"></param>
Procedure TGameState.Initialise(NewImages : TS2DLImageList);
Begin
    Images := NewImages;
    LoadResources();
End;

procedure TGameState.SetKeyboard(const Value: TKeyboard);
begin
  FKeyboard := Value;
end;

Procedure TGameState.Input;
Var
  Event: TSDL_Event;   // Global Event Variable
Begin
  { DONE : Set the Mouse and Keyboard stuff }
  ReturnCode := rt_continue;
  while SDL_PollEvent(@event) > 0 do
  begin
    case event.type_ of
      SDL_VIDEORESIZE:
        Begin
          s2dldraw.windowresize(event.resize.w,event.resize.h);
        End;
      SDL_MOUSEMOTION :
        Begin
          Mouse.X := Event.motion.x;
          Mouse.Y := Event.motion.y;
          S2DLDraw.CursorX := Mouse.X;
          S2DLDraw.CursorY := Mouse.Y;
        End;
      SDL_MOUSEBUTTONDOWN:
        Begin
          If event.button.button = SDL_BUTTON_LEFT then
            Begin
              If Mouse.LeftButton = False then
                Mouse.LeftPressed := True
              Else
                Mouse.LeftPressed := False;
              Mouse.LeftButton := True;
              Mouse.LeftReleased := false;
            End;
          If event.button.button = SDL_BUTTON_RIGHT then
            Begin
              If Mouse.RightButton = False then
                Mouse.RightPressed := True
              Else
                Mouse.RightPressed := False;
              Mouse.RightButton := True;
              Mouse.RightReleased := false;
            End;
        End;
      SDL_MOUSEBUTTONUP:
        Begin
          If event.button.button = SDL_BUTTON_LEFT then
            Begin
              If Mouse.LeftButton = True then
                Mouse.LeftReleased := True
              Else
                Mouse.LeftReleased := False;
              Mouse.LeftButton := false;
              Mouse.LeftPressed := false;
            End;
          If event.button.button = SDL_BUTTON_RIGHT then
            Begin
              If Mouse.RightButton = True then
                Mouse.RightReleased := True
              Else
                Mouse.RightReleased := False;
              Mouse.RightButton := false;
              Mouse.RightPressed := false;
            End;
        End;
      SDL_QUITEV :
        Begin
          ReturnCode := rt_quit;
        End;
      SDL_KEYDOWN:
        begin
          If Keyboard.KeyState[Event.key.keysym.sym] = false then
            Keyboard.KeyPressed[Event.key.keysym.sym] := True
          Else
            Keyboard.KeyPressed[Event.key.keysym.sym] := False;
          Keyboard.KeyState[Event.key.keysym.sym] := True;
          Keyboard.KeyReleased[Event.key.keysym.sym] := False;
        end;
      SDL_KEYUP:
        begin
          If Keyboard.KeyState[Event.key.keysym.sym] = true then
            Keyboard.KeyReleased[Event.key.keysym.sym] := True
          Else
            Keyboard.KeyReleased[Event.key.keysym.sym] := False;
          Keyboard.KeyState[Event.key.keysym.sym] := false;
          Keyboard.KeyPressed[Event.key.keysym.sym] := False;
        end;
    end;
  end;
  DoInput();
End;

/// <summary>
/// Update the objects etc
/// </summary>
Procedure TGameState.Update;
Var
  T, Tick : Integer;
Begin
  Tick := SDL_GetTicks;
  T := Tick - LastTick;
  DoUpdate(T);
  LastTick := Tick;
End;

/// <summary>
/// Load any special resources required by this state.
/// </summary>
Procedure TGameState.LoadResources;
Begin
End;

/// <summary>
/// Call the Draw function
/// Display the Cursor if neccessary.
/// </summary>
Procedure TGameState.Draw;
Begin
    S2DLDraw.Cursor := FCursorOn;
    S2DLDraw.CursorX := Mouse.X;
    S2DLDraw.CursorY := Mouse.Y;
    S2DLDraw.Clear;
    S2DLDraw.SceneStart;
    DoDraw();

    S2DLDraw.SceneEnd;
End;

/// <summary>
/// Update the game state
/// </summary>
Procedure TGameState.DoUpdate(GameTime : LongInt);
Begin
End;

/// <summary>
/// Called when a state is entered for the first time
/// </summary>
Procedure TGameState.DoStartState;
Begin
    // mouse.Clear();
  LastTick := SDL_GetTicks;
  fKeyboard.reset;
End;

/// <summary>
/// Called when the state exists
/// </summary>
Procedure TGameState.DoEndState();
Begin
    if Assigned(OnEndState) then
    Begin
        OnEndState(Self);
    End;
End;

/// <summary>
/// Override to respond to user Input
/// </summary>
Procedure TGameState.DoInput();
Begin
End;

/// <summary>
/// Draw the state onto the screen
/// </summary>
Procedure TGameState.DoDraw();
Begin
End;

procedure TGameState.SetCursor(const Value: TS2DLImage);
begin
  FCursor := Value;
end;

procedure TGameState.SetFont(const Value: TS2DLNeheFont);
begin
  FFont := Value;
end;

procedure TGameState.SetGameTitle(const Value: String);
begin
  FGameTitle := Value;
end;

procedure TGameState.SetActive(const Value: Boolean);
begin
  FActive := Value;
end;

procedure TGameState.SetCursorOn(const Value: Boolean);
begin
  FCursorOn := Value;
  S2DLDraw.Cursor := FCursorOn;
end;

procedure TGameState.SetOnEndState(const Value: TNotifyEvent);
begin
FOnEndState := Value;
end;

/// <summary>
/// Game State Image
/// Displays an Image - closes on Keyboard or Mouse event
/// </summary>

procedure TGameStateImage.DoDraw();
Begin
//inherited DoDraw();
//        BackImage.Draw(0,0);
  BackImage.DrawStretch(0,0,800,600,0);
End;

procedure TGameStateImage.DoInput();
Begin
inherited DoInput();
if (mouse.LeftButton = Pressed) then
Begin
    DoEndState();
End;
End;

procedure TGameStateImage.SetBackImage(const Value: TS2DLImage);
begin
FBackImage := Value;
end;

procedure TGameState.SetMouse(const Value: TMouse);
begin
FMouse := Value;
end;

{ TKeyboard }

constructor TKeyboard.Create;

begin
 reset;
end;

procedure Tkeyboard.reset;
Var
  I : integer;
begin
  For I := SDLK_FIRST to SDLK_EURO do
  Begin
    KeyState[i] := False;
    KeyPressed[i] := False;
    KeyReleased[i] := False;
  End;
end;

{ TGameStateSplash }
constructor TGameStateSplash.Create;
begin
  inherited Create;

  Delay := 1;
end;

procedure TGameStateSplash.SetDelay(const Value: LongInt);
begin
  FDelay := Value;
end;

Procedure TGameStateSplash.DoStartState();
Begin
  Inherited DoStartState();
  DelayTime := -1;
End;

Procedure TGameStateSplash.DoEndState();
Begin
  Active := false;
  Inherited DoEndState();
End;

Procedure TGameStateSplash.DoUpdate(GameTime : LongInt);
Begin
    Inherited DoUpdate(gameTime);
    DelayTime := DelayTime + gameTime;
    if (DelayTime > (Delay * 1000)) then
    Begin
        DoEndState();
    End;
End;

Procedure TGameStateSplash.DoDraw();
Var
  S : Integer;
Begin
    Inherited DoDraw();

    Font.WriteText(200, 0, GameTitle, 1);
    S := (DelayTime div 1000);
    Font.WriteTextStretch(0,0,0.5,'Seconds:'+IntToStr(S));
    Font.WriteTextStretch(0,50,0.5,'FPS:'+IntToStr(S2DLDraw.FPS));
End;

Procedure TGameStateSplash.DoInput();
Begin
    Inherited DoInput();
    if (mouse.LeftPressed) then
    Begin
        DoEndState();
    End;
End;

procedure TGameStateSplash.SetDelayTime(const Value: LongInt);
begin
  FDelayTime := Value;
end;

procedure TGameState.SetLastTick(const Value: Cardinal);
begin
  FLastTick := Value;
end;

End.

